# Reseda version 1.0.0
# ResedaEngine.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of Reseda.
# Reseda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Reseda is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Reseda.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Settings
# 4) Builtins
# 5) Calculate and print values of elements
# 6) Reset and draw  values of random deviates
# 7) Calculate values of set expressions
# 8) Network operations

# 1) Imported modules
# -------------------

from S2MLCore import *

from ResedaModels import *
from ResedaInterface import *
from ResedaModelManager import *
from ResedaInterpreter import *

from ResedaShortestPathCalculator import *
from ResedaPERTCalculator import *
from ResedaMaximumFlowCalculator import *

from MusetCore import *

# 2) Creator
# ----------

class ResedaEngine(S2MLEngine):

  def __init__(self):
    S2MLEngine.__init__(self)
    self._interpreter = None
    self._targetBlock = None
    self._parameters = []
    self._nodes = []
    self._edges = []
    self._variables = []
    self._contextualElement = None
    self._musetCalculator = None
    self._shortestPathCalculator = None
    self._PERTCalculator = None
    self._maximumFlowCalculator = None

# 3) Settings
# -----------

  def CreateTools(self):
    self._modelManager = ResedaModelManager(self)
    self._interpreter = ResedaInterpreter(self)
    self._environmentManager = S2MLEnvironmentManager(self)
    self._modelManager.CreateTools()
    self._interpreter.CreateTools()
    self._environmentManager.CreateTools()
    self._environmentManager.Initialize()
    self._expressionManager = S2MLStochasticExpressionManager(self)
    self._expressionManager.SetCalculateValueOfReferenceExpression(ResedaEngine.CalculateValueOfReferenceExpression)
    self._expressionManager.SetCalculateValueOfElement(ResedaEngine.CalculateValueOfElement)
    self._musetCalculator = MusetCalculator(self)
    self._shortestPathCalculator = ResedaShortestPathCalculator(self)
    self._PERTCalculator = ResedaPERTCalculator(self)
    self._maximumFlowCalculator = ResedaMaximumFlowCalculator(self)
    self._modelManager.ResetModel()

  def GetContextualElement(self):
    return self._contextualElement

  def SetContextualElement(self, element):
    self._contextualElement = element

# 4) Builtins
# -----------

  def ExecuteCommandCompute(self, command):
    what = command.GetArgumentAsIdentifier(0)
    if what=="shortest-path":
      return self._ExecuteCommandComputeShortestPath(command)
    if what=="dates":
      return self._ExecuteCommandComputeDates(command)
    if what=="maximum-flow":
      return self._ExecuteCommandComputeMaximumFlow(command)
    if what=="values":
      return self._ExecuteCommandComputeValues(command, True, True, True)
    if what=="parameters":
      return self._ExecuteCommandComputeValues(command, True, False, False)
    if what=="attributes":
      return self._ExecuteCommandComputeValues(command, False, True, False)
    if what=="properties":
      return self._ExecuteCommandComputeValues(command, False, False, True)
    self.RaiseErrorInCommand(command, "Incorrect argument {0:s}\n".format(what))
    return 1

  # compute shortest-path source-node target-node attribute-name [output=fileName] [mode=(append|write)]
  def _ExecuteCommandComputeShortestPath(self, command):
    sourceNodeName = command.GetArgumentAsIdentifier(1)
    targetNodeName = command.GetArgumentAsIdentifier(2)
    attributeName = command.GetArgumentAsIdentifier(3)
    targetBlock = self.GetValueOfOptionOfCommandAsTargetBlock(command)
    if targetBlock==None:
      self.RaiseError("No target block")
      return None
    matrix = self._shortestPathCalculator.ComputeShortestPathWithDijskraAlgorithm(targetBlock, sourceNodeName, targetNodeName, attributeName)
    if matrix==None:
      return 1
    printer = self._modelManager.GetModelPrinter()
    outputFileName = self.GetValueOfOptionOfCommandAsString(command, "output")
    mode = self.GetValueOfOptionOfCommandAsFileOpeningMode(command, "mode")
    if outputFileName==None:
      error = printer.PrintMatrix(matrix, self._interpreter.GetOutputFile())
    else:
      outputFilePath = self.GetFilePath(outputFileName)
      error = printer.ExportMatrix(matrix, outputFilePath, mode)
    return error

  # compute dates source-node target-node attribute-name [output=fileName] [mode=(append|write)]
  def _ExecuteCommandComputeDates(self, command):
    sourceNodeName = command.GetArgumentAsIdentifier(1)
    targetNodeName = command.GetArgumentAsIdentifier(2)
    attributeName = command.GetArgumentAsIdentifier(3)
    targetBlock = self.GetValueOfOptionOfCommandAsTargetBlock(command)
    if targetBlock==None:
      self.RaiseError("No target block")
      return None
    matrix = self._PERTCalculator.ComputeDatesOfPERTDiagram(targetBlock, sourceNodeName, targetNodeName, attributeName)
    if matrix==None:
      return 1
    printer = self._modelManager.GetModelPrinter()
    outputFileName = self.GetValueOfOptionOfCommandAsString(command, "output")
    mode = self.GetValueOfOptionOfCommandAsFileOpeningMode(command, "mode")
    if outputFileName==None:
      error = printer.PrintMatrix(matrix, self._interpreter.GetOutputFile())
    else:
      outputFilePath = self.GetFilePath(outputFileName)
      error = printer.ExportMatrix(matrix, outputFilePath, mode)
    return error

  # compute maximum-flow source-node target-node attribute-name fileName [mode=(append|write)]
  def _ExecuteCommandComputeMaximumFlow(self, command):
    sourceNodeName = command.GetArgumentAsIdentifier(1)
    targetNodeName = command.GetArgumentAsIdentifier(2)
    attributeName = command.GetArgumentAsIdentifier(3)
    targetBlock = self.GetValueOfOptionOfCommandAsTargetBlock(command)
    if targetBlock==None:
      self.RaiseError("No target block")
      return None
    matrix = self._maximumFlowCalculator.ComputeMaximumFlowWithFordFulkersonAlgorithm(targetBlock, sourceNodeName, targetNodeName, attributeName)
    if matrix==None:
      return 1
    printer = self._modelManager.GetModelPrinter()
    outputFileName = self.GetValueOfOptionOfCommandAsString(command, "output")
    mode = self.GetValueOfOptionOfCommandAsFileOpeningMode(command, "mode")
    if outputFileName==None:
      error = printer.PrintMatrix(matrix, self._interpreter.GetOutputFile())
    else:
      outputFilePath = self.GetFilePath(outputFileName)
      error = printer.ExportMatrix(matrix, outputFilePath, mode)
    return error

  def _ExecuteCommandComputeValues(self, command, printParameters, printAttributes, printProperties):
    targetBlock = self.GetValueOfOptionOfCommandAsTargetBlock(command)
    if targetBlock==None:
      self.RaiseError("No target block")
      return None
    missionTimes = self.GetValueOfOptionOfCommandAsListOfReals(command, "mission-time", [0.0])
    if missionTimes==None:
      return None
    outputFileName = self.GetValueOfOptionOfCommandAsString(command, "output")
    mode = self.GetValueOfOptionOfCommandAsFileOpeningMode(command, "mode")
    if outputFileName==None:
       outputFile = self._interpreter.GetOutputFile()
    else:
      try:
        outputFilePath = self.GetFilePath(outputFileName)
        outputFile = open(outputFilePath, mode)
      except:
        self.RaiseError('Unable to open file "{0:s}"\n'.format(self._outputFileName))
        return 1
    self.CalculateAndPrintValuesOfElements(targetBlock, missionTimes, printParameters, printAttributes, printProperties, outputFile)
    if outputFileName!=None:
      outputFile.close()
    return 0

# 5) Methods to calculate and print values of elements
# ----------------------------------------------------

  def CalculateAndPrintValuesOfElements(self, targetBlock, missionTimes, printParameters, printAttributes, printProperties, outputFile):
    self._targetBlock = targetBlock
    self._parameters.clear()
    self._nodes.clear()
    self._edges.clear()
    self._variables.clear()
    self._CollectModelElements(targetBlock)
    self._ResetValuesOfElements()
    for missionTime in missionTimes:
      self._expressionManager.SetMissionTime(missionTime)
      self._ResetAbscissasOfRandomDeviatesOfElements()
      self._ResetValuesOfElements()
      self._CalculateValuesOfElements()
      self._modelManager.GetModelPrinter().PrintValuesOfElements(targetBlock, printParameters, printAttributes, printProperties, outputFile)
    self._targetBlock = None
    return 0

  def _CollectModelElements(self, block):
    for element in block.GetElements():
      if element.GetType()==ResedaType.BLOCK:
        self._CollectModelElements(element)
      elif element.GetType()==ResedaType.PARAMETER:
        self._parameters.append(element)
      elif element.GetType()==ResedaType.NODE:
        self._nodes.append(element)
      elif element.GetType()==ResedaType.EDGE:
        self._edges.append(element)
      elif element.GetType()==ResedaType.ARC:
        self._edges.append(element)
      elif element.GetType()==ResedaType.VARIABLE:
        self._variables.append(element)

  @staticmethod
  def CalculateValueOfReferenceExpression(expressionManager, expression):
    engine = expressionManager.GetEngine()
    element = engine.GetContextualElement()
    if element==None:
      return S2MLStochasticExpressionManager.CalculateValueOfReferenceExpression(engine._expressionManager, expression)
    path = expression.GetReferencePath()
    if not path.IsIdentifier():
      return S2MLConstant.NewNone()
    name = path.GetFirstIdentifier()
    attributeExpression = element.GetAttribute(name)
    if attributeExpression==None:
      return S2MLConstant.NewNone()
    return attributeExpression.GetValue()

  @staticmethod
  def CalculateValueOfElement(expressionManager, element):
    if element.GetType()==ResedaType.PARAMETER:
      return expressionManager.CalculateValueOfParameter(element)
    return S2MLConstant.NewNone()

  def _ResetValuesOfElements(self):
    for parameter in self._parameters:
      parameter.ResetValue()
    for variable in self._variables:
      variable.ResetValue()

  def _CalculateValuesOfElements(self):
    for parameter in self._parameters:
      self._expressionManager.CalculateValueOfParameter(parameter)
    for node in self._nodes:
      self._CalculateValuesOfAttributesOfElement(node)
    for edge in self._edges:
      self._CalculateValuesOfAttributesOfElement(edge)
    for variable in self._variables:
      musetSet = MusetSet.NewEmptySet()
      variable.SetValue(musetSet)
    fixpoint = False
    while not fixpoint:
      fixpoint = True
      for variable in self._variables:
        modified = self._CalculateValueOfVariable(variable)
        fixpoint = fixpoint and not modified

  def _CalculateValuesOfAttributesOfElement(self, element):
    for value in element.GetAttributeValues():
      self._expressionManager.CalculateValueOfExpression(value)

  def _CalculateValueOfVariable(self, variable):
    previousMusetSet = variable.GetValue()
    parentBlock = variable.GetContainer()
    definition = variable.GetDefinition()
    newMusetSet = self._CalculateValueOfSetExpression(definition, parentBlock)
    variable.SetValue(newMusetSet)
    return not newMusetSet.IsEqual(previousMusetSet)

# 6) Reset and draw  values of random deviates
# --------------------------------------------

  def _ResetAbscissasOfRandomDeviatesOfElements(self):
    for parameter in self._parameters:
      definition = parameter.GetDefinition()
      self._expressionManager.ResetAbscissasOfRandomDeviatesInExpression(definition)
    for node in self._nodes:
      self._ResetAbscissasOfRandomDeviatesOfAttributesOfElement(node)
    for edge in self._edges:
      self._ResetAbscissasOfRandomDeviatesOfAttributesOfElement(edge)
    for variable in self._variables:
      self._ResetAbscissasOfRandomDeviatesOfVariable(variable)

  def _ResetAbscissasOfRandomDeviatesOfAttributesOfElement(self, element):
    for value in element.GetAttributeValues():
      self._expressionManager.ResetAbscissasOfRandomDeviatesInExpression(value)

  def _ResetAbscissasOfRandomDeviatesOfVariable(self, variable):
    pass

  def _DrawAbscissasOfRandomDeviatesOfElements(self):
    for parameter in self._parameters:
      definition = parameter.GetDefinition()
      self._expressionManager.DrawAbscissasOfRandomDeviatesInExpression(definition)
    for node in self._nodes:
      self._DrawAbscissasOfRandomDeviatesOfAttributesOfElement(node)
    for edge in self._edges:
      self._DrawAbscissasOfRandomDeviatesOfAttributesOfElement(edge)
    for variable in self._variables:
      self._DrawAbscissasOfRandomDeviatesOfVariable(variable)

  def _DrawAbscissasOfRandomDeviatesOfAttributesOfElement(self, element):
    for value in element.GetAttributeValues():
      self._expressionManager.DrawAbscissasOfRandomDeviatesInExpression(value)

  def _DrawAbscissasOfRandomDeviatesOfVariable(self, variable):
    pass

# 7) Calculate values of set expressions
# --------------------------------------

  def _CalculateValueOfSetExpression(self, expression, parentBlock):
    if expression.GetType()==ResedaSetExpression.ELEMENT_REFERENCE:
      return self._CalculateValueOfSetExpressionReference(expression, parentBlock)
    if expression.GetType()==ResedaSetExpression.EXTENSIONAL_SET:
      return self._CalculateValueOfSetExpressionExtensionalSet(expression, parentBlock)
    if expression.GetType()==ResedaSetExpression.UNION:
      return self._CalculateValueOfSetExpressionAssociativeOperation(expression, parentBlock, self._musetCalculator.BuildUnion)
    if expression.GetType()==ResedaSetExpression.INTERSECTION:
      return self._CalculateValueOfSetExpressionAssociativeOperation(expression, parentBlock, self._musetCalculator.BuildIntersection)
    if expression.GetType()==ResedaSetExpression.SET_DIFFERENCE:
      return self._CalculateValueOfSetExpressionAssociativeOperation(expression, parentBlock, self._musetCalculator.BuildDifference)
    if expression.GetType()==ResedaSetExpression.SELECTION:
      return self._CalculateValueOfSetExpressionSelection(expression, parentBlock)
    if expression.GetType()==ResedaSetExpression.PARENT_BLOCKS:
      return self._CalculateValueOfSetExpressionNetworkUnaryOperation(expression, parentBlock, self._musetCalculator.CollectParentBlocksOfElement)
    if expression.GetType()==ResedaSetExpression.ANCESTOR_BLOCKS:
      return self._CalculateValueOfSetExpressionNetworkUnaryOperation(expression, parentBlock, self._musetCalculator.CollectAncestorBlocksOfElement)
    if expression.GetType()==ResedaSetExpression.CHILD_BLOCKS:
      return self._CalculateValueOfSetExpressionNetworkUnaryOperation(expression, parentBlock, self._musetCalculator.CollectChildBlocksOfElement)
    if expression.GetType()==ResedaSetExpression.DESCENDANT_BLOCKS:
      return self._CalculateValueOfSetExpressionNetworkUnaryOperation(expression, parentBlock, self._musetCalculator.CollectDescendantBlocksOfElement)
    if expression.GetType()==ResedaSetExpression.CHILD_NODES:
      return self._CalculateValueOfSetExpressionNetworkUnaryOperation(expression, parentBlock, self._musetCalculator.CollectChildNodesOfElement)
    if expression.GetType()==ResedaSetExpression.DESCENDANT_NODES:
      return self._CalculateValueOfSetExpressionNetworkUnaryOperation(expression, parentBlock, self._musetCalculator.CollectDescendantNodesOfElement)
    if expression.GetType()==ResedaSetExpression.CHILD_EDGES:
      return self._CalculateValueOfSetExpressionNetworkUnaryOperation(expression, parentBlock, self._musetCalculator.CollectChildEdgesOfElement)
    if expression.GetType()==ResedaSetExpression.DESCENDANT_EDGES:
      return self._CalculateValueOfSetExpressionNetworkUnaryOperation(expression, parentBlock, self._musetCalculator.CollectDescendantEdgesOfElement)
    if expression.GetType()==ResedaSetExpression.SOURCE_NODES:
      return self._CalculateValueOfSetExpressionNetworkUnaryOperation(expression, parentBlock, self._musetCalculator.CollectSourceNodesOfElement)
    if expression.GetType()==ResedaSetExpression.TARGET_NODES:
      return self._CalculateValueOfSetExpressionNetworkUnaryOperation(expression, parentBlock, self._musetCalculator.CollectTargetNodesOfElement)
    if expression.GetType()==ResedaSetExpression.IN_EDGES:
      return self._CalculateValueOfSetExpressionNetworkUnaryOperation(expression, parentBlock, self._musetCalculator.CollectInEdgesOfElement)
    if expression.GetType()==ResedaSetExpression.OUT_EDGES:
      return self._CalculateValueOfSetExpressionNetworkUnaryOperation(expression, parentBlock, self._musetCalculator.CollectOutEdgesOfElement)
    if expression.GetType()==ResedaSetExpression.REACHABLE_NODES:
      return self._CalculateValueOfSetExpressionNetworkReachableNodesOperation(expression, parentBlock, self._musetCalculator.CollectReachableNodesOfElement)
    if expression.GetType()==ResedaSetExpression.COREACHABLE_NODES:
      return self._CalculateValueOfSetExpressionNetworkReachableNodesOperation(expression, parentBlock, self._musetCalculator.CollectCoReachableNodesOfElement)
    return MusetSet.NewEmptySet()

  def _CalculateValueOfSetExpressionReference(self, expression, parentBlock):
    path = expression.GetReferencePath()
    element = parentBlock.AccessNamedElement(path)
    if element==None:
      return MusetSet.NewEmptySet()
    if element.GetType()==ResedaType.BLOCK:
      return MusetElement.NewBlock(element)
    if element.GetType()==ResedaType.NODE:
      return MusetElement.NewNode(element)
    if element.GetType()==ResedaType.EDGE:
      return MusetElement.NewEdge(element)
    if element.GetType()==ResedaType.VARIABLE:
      sourceMusetSet = element.GetValue()
      targetMusetSet = MusetSet.Copy(sourceMusetSet)
      return targetMusetSet
    return MusetSet.NewEmptySet()

  def _CalculateValueOfSetExpressionExtensionalSet(self, expression, parentBlock):
    musetSet = MusetSet.NewEmptySet()
    for child in expression.GetChildren():
      musetChildValue = self._CalculateValueOfSetExpression(child, parentBlock)
      musetSet.InsertElement(musetChildValue)
    return musetSet

  def _CalculateValueOfSetExpressionAssociativeOperation(self, expression, parentBlock, operation):
    musetSet = None
    for child in expression.GetChildren():
      musetChildValue = self._CalculateValueOfSetExpression(child, parentBlock)
      if musetSet==None:
        musetSet = musetChildValue
      else:
        musetSet = operation(musetSet, musetChildValue)
    if musetSet==None:
      return MusetSet.NewEmptySet()
    return musetSet

  def _CalculateValueOfSetExpressionSelection(self, expression, parentBlock):
    child = expression.GetChild(0)
    musetChildSet = self._CalculateValueOfSetExpression(child, parentBlock)
    if not musetChildSet.IsSet():
      return MusetSet.NewEmptySet()
    testOnAttributes = expression.GetTestOnAttributes(0)
    musetSet = MusetSet.NewEmptySet()
    for musetElement in musetChildSet.GetElements():
      element = musetElement.GetNetworkElement()
      if element==None:
        continue
      satisfied = self.CalculateValueOfTestOnAttributes(element, testOnAttributes)
      if satisfied:
        musetSet.AppendElement(musetElement)
    return musetSet

  def CalculateValueOfTestOnAttributes(self, element, testOnAttributes):
    self.SetContextualElement(element)
    value = self._expressionManager.CalculateValueOfExpression(testOnAttributes)
    self.SetContextualElement(None)
    return value.IsTrue()

# 8) Network operations
# ---------------------

  def _CalculateValueOfSetExpressionNetworkUnaryOperation(self, expression, parentBlock, operation):
    child = expression.GetChild(0)
    musetChildElement = self._CalculateValueOfSetExpression(child, parentBlock)
    musetSet = MusetSet.NewEmptySet()
    self._musetCalculator.SetTargetBlock(self._targetBlock)
    operation(musetChildElement, musetSet)
    return musetSet

  def _CalculateValueOfSetExpressionNetworkReachableNodesOperation(self, expression, parentBlock, operation):
    child = expression.GetChild(0)
    musetChildElement = self._CalculateValueOfSetExpression(child, parentBlock)
    musetSet = MusetSet.NewEmptySet()
    self._musetCalculator.SetTargetBlock(self._targetBlock)
    testOnAttributesOfNodes = expression.GetTestOnAttributes(0)
    testOnAttributesOfEdges = expression.GetTestOnAttributes(1)
    operation(musetChildElement, musetSet, testOnAttributesOfNodes, testOnAttributesOfEdges)
    return musetSet

