# Reseda version 1.0.0
# ResedaModelRewriter.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of Reseda.
# Reseda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Reseda is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Reseda.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Post-processing
# 4) Definition of references

# 1) Imported modules
# -------------------

from S2MLCore import *

from ResedaModels import *

# 2) Creator
# ----------

class ResedaModelRewriter(S2MLModelRewriter):
  UNDEFINED = -1

  def __init__(self, engine):
    S2MLModelRewriter.__init__(self, engine)
    self._currentBlock = None
    self._errors = 0

# 3) Post-processing
# ------------------

  def PostProcessModel(self, model):
    self._errors = 0
    for element in model.GetElements():
      if element.GetType()==ResedaType.BLOCK:
        self.PostProcessBlock(element)
        if self._errors!=0:
          return self._errors
    return self._errors

  def PostProcessBlock(self, block):
    self.DefineReferencesInBlock(block)

# 4) Definition of references
# ---------------------------

  def DefineReferencesInBlock(self, block):
    previousBlock = self._currentBlock
    self._currentBlock = block
    for element in block.GetElements():
      if element.GetType()==ResedaType.BLOCK:
        self.DefineReferencesInBlock(element)
      elif element.GetType()==ResedaType.PARAMETER:
        self.DefineReferencesInParameterDeclaration(element)
      elif element.GetType()==ResedaType.NODE:
        self.DefineReferencesInNodeDeclaration(element)
      elif element.GetType()==ResedaType.EDGE:
        self.DefineReferencesInEdgeDeclaration(element)
      elif element.GetType()==ResedaType.ARC:
        self.DefineReferencesInEdgeDeclaration(element)
    self._currentBlock = previousBlock

  def DefineReferencesInParameterDeclaration(self, parameter):
    definition = parameter.GetDefinition()
    self.DefineReferencesInStochasticExpression(definition)

  def DefineReferencesInNodeDeclaration(self, node):
    for name in node.GetAttributeNames():
      expression = node.GetAttribute(name)
      self.DefineReferencesInStochasticExpression(expression)

  def DefineReferencesInEdgeDeclaration(self, edge):
    for name in edge.GetAttributeNames():
      expression = edge.GetAttribute(name)
      self.DefineReferencesInStochasticExpression(expression)

  def DefineReferencesInStochasticExpression(self, expression):
    if expression.GetType()==ResedaType.REFERENCE_EXPRESSION:
      path = expression.GetReferencePath()
      element = self._currentBlock.AccessNamedElement(path)
      if element==None:
        self.RaiseError("In block {0:s}, undefined reference {1:s}".format(self._currentBlock.GetIdentifier(), path.ToString()))
        self._errors += 1
        return
      expression.SetReferredElement(element)
      return
    if expression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      return
    for child in expression.GetChildren():
      self.DefineReferencesInStochasticExpression(child)
