# Reseda version 1.0.0
# ResedaPERTCalculator.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of Reseda.
# Reseda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Reseda is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Reseda.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) PERT method

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLCore import *
from ResedaModels import *

# 2) Creator
# ----------

class ResedaPERTCalculator:
  DURATION = "@DURATION"
  EARLY_DATE = "@EARLY_DATE"
  LATE_DATE = "@LATE_DATE"

  def __init__(self, engine):
    from ResedaEngine import ResedaEngine
    self._engine = engine
    self._expressionManager = self._engine.GetExpressionManager()
    self._targetBlock = None
    self._sourceNode = None
    self._targetNode = None
    self._matrix = None

  def RaiseError(self, message):
    self._engine.RaiseError(message)

# 3) PERT method
# --------------

  def ComputeDatesOfPERTDiagram(self, block, sourceNodeName, targetNodeName, attributeName):
    self._targetBlock = block
    if self._targetBlock==None:
      self.RaiseError("No target block")
      return None
    sourceNodePath = S2MLPath.FromString(sourceNodeName)
    self._sourceNode = self._targetBlock.AccessNode(sourceNodePath)
    if self._sourceNode==None:
      self.RaiseError("No node %s" % sourceNodeName)
      return None
    targetNodePath = S2MLPath.FromString(targetNodeName)
    self._targetNode = self._targetBlock.AccessNode(targetNodePath)
    if self._targetNode==None:
      self.RaiseError("No node %s" % targetNodeName)
      return None
    error = self.BuildAndInitializeGraphToComputeDates(attributeName)
    if error:
      return None
    self.CalculateEarlyDates()
    self.CalculateLateDates()
    return self.CreateResultMatrixForDates()

  def BuildAndInitializeGraphToComputeDates(self, attributeName):
    self._targetBlock.BuildGraph()
    self._expressionManager.ResetValuesOfParameters(self._targetBlock)
    for edge in self._targetBlock.GetEdges():
      expression = edge.GetAttribute(attributeName)
      if expression==None:
        self.RaiseError("Attribute %s missing in edge %s." % (attributeName, edge.GetIdentifier()))
        return 1
      else:
        value = self._expressionManager.CalculateValueOfExpression(expression)
        if not value.IsNumber():
          self.RaiseError("Attribute %s of edge %s is not a number" % (attributeName, edge.GetIdentifier()))
          return 1
        duration = value.ToFloat()
        if duration<0:
          self.RaiseError("Attribute %s of edge %s has a negative value" % (attributeName, edge.GetIdentifier()))
          return 1
        edge.SetAttribute(ResedaPERTCalculator.DURATION, duration)
    for node in self._targetBlock.GetNodes():
      node.SetAttribute(ResedaPERTCalculator.EARLY_DATE, math.inf)
      node.SetAttribute(ResedaPERTCalculator.LATE_DATE, -math.inf)
    return 0

  def CalculateEarlyDates(self):
    self._sourceNode.SetAttribute(ResedaPERTCalculator.EARLY_DATE, 0.0)
    while True:
      node = self.SelectEarlyDateCandidate()
      if node==None:
        break

  def SelectEarlyDateCandidate(self):
    for node in self._targetBlock.GetNodes():
      nodeEarlyDate = node.GetAttribute(ResedaPERTCalculator.EARLY_DATE)
      if nodeEarlyDate!=math.inf:
        continue
      maximumEarlyDate = 0.0
      allCalculated = True
      for edge in node.GetInEdges():
        edgeDuration = edge.GetAttribute(ResedaPERTCalculator.DURATION)
        sourceNode = edge.GetSourceNode()
        sourceNodeEarlyDate = sourceNode.GetAttribute(ResedaPERTCalculator.EARLY_DATE)
        if sourceNodeEarlyDate==math.inf:
          allCalculated = False
          break
        if sourceNodeEarlyDate + edgeDuration>maximumEarlyDate:
          maximumEarlyDate = sourceNodeEarlyDate + edgeDuration
      if allCalculated:
        node.SetAttribute(ResedaPERTCalculator.EARLY_DATE, maximumEarlyDate)
        return node
    return None

  def CalculateLateDates(self):
    self._targetNode.SetAttribute(ResedaPERTCalculator.LATE_DATE, self._targetNode.GetAttribute(ResedaPERTCalculator.EARLY_DATE))
    while True:
      node = self.SelectLateDateCandidate()
      if node==None:
        break

  def SelectLateDateCandidate(self):
    for node in self._targetBlock.GetNodes():
      nodeLateDate = node.GetAttribute(ResedaPERTCalculator.LATE_DATE)
      if nodeLateDate!=-math.inf:
        continue
      minimumLateDate = math.inf
      allCalculated = True
      for edge in node.GetOutEdges():
        edgeDuration = edge.GetAttribute(ResedaPERTCalculator.DURATION)
        targetNode = edge.GetTargetNode()
        targetNodeLateDate = targetNode.GetAttribute(ResedaPERTCalculator.LATE_DATE)
        if targetNodeLateDate==-math.inf:
          allCalculated = False
          break
        elif targetNodeLateDate - edgeDuration<minimumLateDate:
          minimumLateDate = targetNodeLateDate - edgeDuration
      if allCalculated:
        node.SetAttribute(ResedaPERTCalculator.LATE_DATE, minimumLateDate)
        return node
    return None

  def GetCriticalPath(self, node, path):
    if node==self._targetNode:
      return True
    for edge in node.GetOutEdges():
      oppositeNode = edge.GetOppositeNode(node)
      if oppositeNode.GetDistance()==node.GetDistance() + edge.GetLength():
        path.append(edge)
        found = self.GetShortestPath(oppositeNode, path)
        if found:
          return True
        path.pop()
    return False

  def CreateResultMatrixForDates(self):
    matrix = S2MLMatrix()
    row = matrix.AppendRow()
    row.Put(0, S2MLConstant.NewString("Model"))
    row.Put(1, S2MLConstant.NewString(self._targetBlock.GetIdentifier()))
    row = matrix.AppendRow()
    row.Put(0, S2MLConstant.NewString("source"))
    row.Put(1, S2MLConstant.NewString(self._sourceNode.GetIdentifier()))
    row = matrix.AppendRow()
    row.Put(0, S2MLConstant.NewString("target"))
    row.Put(1, S2MLConstant.NewString(self._targetNode.GetIdentifier()))
    row = matrix.AppendRow()
    row.Put(0, S2MLConstant.NewString("node"))
    row.Put(1, S2MLConstant.NewString("early-date"))
    row.Put(2, S2MLConstant.NewString("late-date"))
    for node in self._targetBlock.GetNodes():
      row = matrix.AppendRow()
      row.Put(0, S2MLConstant.NewString(node.GetIdentifier()))
      row.Put(1, S2MLConstant.NewReal(node.GetAttribute(ResedaPERTCalculator.EARLY_DATE)))
      row.Put(2, S2MLConstant.NewReal(node.GetAttribute(ResedaPERTCalculator.LATE_DATE)))
    return matrix


