# Reseda version 1.0.0
# ResedaWalker.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of Reseda.
# Reseda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Reseda is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Reseda.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Walks
# 4) Preprocessing

# 1) Imported modules
# -------------------

import time

from S2MLCore import *
from ResedaModels import *

# 2) Creator
# ----------

class ResedaWalker:

  def __init__(self, engine):
    from ResedaEngine import ResedaEngine
    self._engine = engine
    self._expressionManager = self._engine.GetExpressionManager()
    self._expressionManager.SetCalculateValueOfElement(ResedaWalker.CalculateValueOfElement)
    self._mainBlock = None
    self._missionTime = 0
    self._parameters = []
    self._nodes = []
    self._edges = []

  def RaiseError(self, message):
    self._engine.RaiseError(message)

# 3) Walks
# --------

  def GetMainBlock(self):
    return self._mainBlock

  def SetMainBlock(self, block):
    self._mainBlock = block

  def GetMissionTime(self):
    return self._missionTime

  def SetMissionTime(self, missionTime):
    self._missionTime = missionTime

  def StartWalk(self):
    if self._mainBlock==None:
      self.RaiseError("Main block undefined")
      return
    self.CalculateValuesOfElements()

  def TerminateWalk(self):
    pass

  def GetOutputFileName(self):
    return self._outputFileName

  def SetOutputFileName(self, outputFileName):
    self._outputFileName = outputFileName

  def GetOutputFileOpeningMode(self):
    return self._outputFileOpeningMode

  def SetOutputFileOpeningMode(self, openingMode):
    self._outputFileOpeningMode = openingMode

# 4) Preprocessing
# ----------------

  def CalculateValuesOfElements(self):
    if self._mainBlock==None:
      return
    self._CollectModelElements(self._mainBlock)
    self._ResetValuesOfElements()
    self._expressionManager.SetMissionTime(self._missionTime)
    self._ResetAbscissasOfRandomDeviatesOfElements()
    self._ResetValuesOfElements()
    self._CalculateValuesOfElements()

  def _CollectModelElements(self, block):
    self._parameters.clear()
    self._nodes.clear()
    self._edges.clear()
    for element in block.GetElements():
      if element.GetType()==ResedaType.BLOCK:
        self._CollectModelElements(element)
      elif element.GetType()==ResedaType.PARAMETER:
        self._parameters.append(element)
      elif element.GetType()==ResedaType.NODE:
        self._nodes.append(element)
      elif element.GetType()==ResedaType.EDGE:
        self._edges.append(element)
      elif element.GetType()==ResedaType.ARC:
        self._edges.append(element)

  @staticmethod
  def CalculateValueOfElement(expressionManager, element):
    if element.GetType()==ResedaType.PARAMETER:
      return expressionManager.CalculateValueOfParameter(element)
    return S2MLConstant.NewNone()

  def _ResetValuesOfElements(self):
    for parameter in self._parameters:
      parameter.ResetValue()

  def _CalculateValuesOfElements(self):
    for parameter in self._parameters:
      self._expressionManager.CalculateValueOfParameter(parameter)
    for node in self._nodes:
      self._CalculateValuesOfAttributesOfElement(node)
    for edge in self._edges:
      self._CalculateValuesOfAttributesOfElement(edge)

  def _CalculateValuesOfAttributesOfElement(self, element):
    for value in element.GetAttributeValues():
      self._expressionManager.CalculateValueOfExpression(value)

  def _ResetAbscissasOfRandomDeviatesOfElements(self):
    for parameter in self._parameters:
      definition = parameter.GetDefinition()
      self._expressionManager.ResetAbscissasOfRandomDeviatesInExpression(definition)
    for node in self._nodes:
      self._ResetAbscissasOfRandomDeviatesOfAttributesOfElement(node)
    for edge in self._edges:
      self._ResetAbscissasOfRandomDeviatesOfAttributesOfElement(edge)

  def _ResetAbscissasOfRandomDeviatesOfAttributesOfElement(self, element):
    for value in element.GetAttributeValues():
      self._expressionManager.ResetAbscissasOfRandomDeviatesInExpression(value)

  def _DrawAbscissasOfRandomDeviatesOfElements(self):
    for parameter in self._parameters:
      definition = parameter.GetDefinition()
      self._expressionManager.DrawAbscissasOfRandomDeviatesInExpression(definition)
    for node in self._nodes:
      self._DrawAbscissasOfRandomDeviatesOfAttributesOfElement(node)
    for edge in self._edges:
      self._DrawtAbscissasOfRandomDeviatesOfAttributesOfElement(edge)

  def _DrawAbscissasOfRandomDeviatesOfAttributesOfElement(self, element):
    for value in element.GetAttributeValues():
      self._expressionManager.DrawAbscissasOfRandomDeviatesInExpression(value)
