# S2MLCore version 1.2.0
# S2MLScriptParser.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Methods to import scripts
# 4) Tokenizer
# 5) Lexer

# 1) Imported modules
# -------------------

import re

from S2MLManagedValue import *
from S2MLCommand import *

# 2) Creator
# ----------

class S2MLScriptParser:
  def __init__(self, engine):
    from S2MLEngine import S2MLEngine
    self._engine = engine
    self._errors = 0
    self._script = None
    self._inputFileName = None
    self._inputFile = None
    self._line = None
    self._lineNumber = 0
    self._tokens = []
    self._listLevel = 0
    self._tokenIndex = 0
    self._managedValues = []
    self._managedValueIndex = 0

  def _RaiseError(self, message):
    self._errors += 1
    self._engine.RaiseError(message)

# 3) Methods to import scripts
# ----------------------------

  def ImportScript(self, fileName):
    filePath = self._engine.GetFilePath(fileName)
    try:
      file = open(filePath, "r")
    except:
      self._RaiseError('Unable to open file "%s"\n' % filePath)
      return None
    script = self.ParseScript(filePath.as_posix(), file)
    file.close()
    if self._errors!=0:
      self._errors = 0
      return None
    return script

  def ParseScript(self, fileName, file):
    previousScript = self._script
    previousInputFileName = self._inputFileName
    previousFile = self._inputFile
    previousLine = self._line
    previousLineNumber = self._lineNumber
    self._script = S2MLCommand.NewScript()
    self._inputFileName = fileName
    self._inputFile = file
    self._line = None
    self._lineNumber = 0
    self._ParseCommands()
    script = self._script
    self._script = previousScript
    self._inputFileName = previousInputFileName
    self._inputFile = previousFile
    self._line = previousLine
    self._lineNumber = previousLineNumber
    return script

  def ParseCommand(self, commandText):
    previousFile = self._inputFile
    previousLine = self._line
    previousLineNumber = self._lineNumber
    self._inputFile = None
    self._line = commandText
    self._lineNumber = 0
    self._ParseCommandLineManagedValues()
    if self._managedValues!=[]:
      self._ParseCommand()
    self._inputFile = previousFile
    self._line = previousLine
    self._lineNumber = previousLineNumber

  def _ParseCommands(self):
    from S2MLEngine import S2MLEngine
    for self._line in self._inputFile:
      if self._engine.GetExecutionStatus()!=S2MLEngine.ACTIVE:
        break
      self._ParseCommandLineManagedValues()
      if self._managedValues!=[]:
        self._ParseCommand()

  def _ParseCommand(self):
    if self._managedValues==[]:
      return
    lineNumber = self._lineNumber
    managedValue = self.GetManagedValue()
    if managedValue==None or not managedValue.IsIdentifier():
      self.RaiseError("line {0:d}\nUnexpected value\n".format(self._lineNumber))
      return
    command = S2MLCommand.NewBuiltIn(managedValue.GetValueOfIdentifier())
    self._script.AppendCommand(command)
    command.SetFileName(self._inputFileName)
    command.SetLineNumber(lineNumber)
    while True:
      argument = self.LookUpManagedValue()
      if argument==None:
        break
      argument = self.GetManagedValue()
      if not argument.IsIdentifier():
        command.AppendArgument(argument)
        continue
      symbol = self.LookUpManagedValue()
      if symbol==None:
        command.AppendArgument(argument)
        break
      if not symbol.IsGivenSymbol("="):
        command.AppendArgument(argument)
        continue
      symbol = self.GetManagedValue()
      value = self.GetManagedValue()
      if value==None:
        self.RaiseError("line {0:d}\nValue of option {1:s} missing\n".format(self._lineNumber, argument.GetValueOfIdentifier()))
        return
      command.SetOption(argument.GetValueOfIdentifier(), value)

# 4) Tokenizer
# ------------

  def _ParseCommandLineTokens(self):
    self._lineNumber += 1
    self._line.rstrip("\r\n")
    self._line.rstrip("\n")
    self._tokens.clear()
    self._tokenIndex = 0
    state = 1
    column = 0
    while True:
      if state==1:
        if column>=len(self._line):
          if self._listLevel==0:
            break
          if self._inputFile==None:
            return
          self._line = self._inputFile.readline()
          self._lineNumber += 1
          self._line.rstrip("\r\n")
          self._line.rstrip("\n")
          column = 0
        else:
          c = self._line[column]
          if c.isspace():
            pass
          elif c=="#":
            break
          elif c=="=":
            self._tokens.append("=")
          elif c==",":
            self._tokens.append(",")
          elif c=="[":
            self._tokens.append("[")
            self._listLevel += 1
          elif c=="]":
            self._tokens.append("]")
            self._listLevel -= 1
          elif c=="'":
            start = column+1
            state = 3
          elif c=='"':
            start = column+1
            state = 4
          elif c=="\\":
            if self._inputFile==None:
              return
            self._line = self._inputFile.readline()
            self._lineNumber += 1
            self._line.rstrip("\r\n")
            self._line.rstrip("\n")
            column = 0
          else:
            start = column
            state = 2
          column += 1
      elif state==2:
        if column>=len(self._line):
          token = self._line[start:column]
          self._tokens.append(token)
          break
        else:
          c = self._line[column]
          if c.isspace() or c=="#" or c=="=" or c=="'" or c=='"' or c=="," or c=="[" or c=="]":
            token = self._line[start:column]
            self._tokens.append(token)
            state = 1
          else:
            column += 1
      elif state==3:
        if column>=len(self._line):
          self._RaiseError("Line %d, unterminated string" % self._lineNumber)
          self._tokens = []
          return
        else:
          c = self._line[column]
          if c=="'":
            token = self._line[start:column]
            self._tokens.append('"' + token + '"')
            state = 1
          column += 1
      elif state==4:
        if column>=len(self._line):
          self._RaiseError("Line %d, unterminated string" % self._lineNumber)
          self._tokens = []
          return
        else:
          c = self._line[column]
          if c=='"':
            token = self._line[start:column]
            self._tokens.append('"' + token + '"')
            state = 1
          column += 1

  def LookUpToken(self):
    if self._tokenIndex>=len(self._tokens):
      return None
    return self._tokens[self._tokenIndex]

  def GetToken(self):
    if self._tokenIndex>=len(self._tokens):
      return None
    token = self._tokens[self._tokenIndex]
    self._tokenIndex += 1
    return token

  def PutBackToken(self):
    self._tokenIndex -= 1

# 5) Lexer
# --------

  def _ParseCommandLineManagedValues(self):
    self._managedValues.clear()
    self._managedValueIndex = 0
    self._ParseCommandLineTokens()
    self._tokenIndex = 0
    while True:
      managedValue = self._ParseCommandLineManagedValue()
      if managedValue==None:
        break
      self._managedValues.append(managedValue)

  def _ParseCommandLineManagedValue(self):
    token = self.GetToken()
    if token==None:
      return None
    if token=="=":
      return S2MLManagedValue.NewSymbol("=")
    if token=="false":
      return S2MLManagedValue.NewBoolean(False)
    if token=="true":
      return S2MLManagedValue.NewBoolean(True)
    if self.IsInt(token):
      return S2MLManagedValue.NewInteger(int(token))
    if self.IsFloat(token):
      return S2MLManagedValue.NewReal(float(token))
    if self.IsIdentifier(token):
      return S2MLManagedValue.NewIdentifier(token)
    if self.IsString(token):
      return S2MLManagedValue.NewString(token[1:-1])
    if token=="[":
      self.PutBackToken()
      return self._ParseCommandLineManagedList()
    self._RaiseError("Line {0:d}. Invalid token {1:s}".format(self._lineNumber, token))

  def IsInt(self, string):
    try:
      result = int(string)
    except:
      return False
    return True

  def IsFloat(self, string):
    try:
      result = float(string)
    except:
      return False
    return True

  def IsIdentifier(self, string):
    return re.match(r"[a-zA-Z-][a-zA-Z0-9-_]*", string)

  def IsString(self, string):
    return re.match(r'"[^"]*"', string)

  def _ParseCommandLineManagedList(self):
    token = self.GetToken()
    if token!="[":
      return None
    managedList = S2MLManagedValue.NewList()
    while True:
      managedValue = self._ParseCommandLineManagedValue()
      if managedValue==None:
        return None
      managedList.AppendValue(managedValue)
      token = self.GetToken()
      if token==None:
        self._RaiseError("Line {0:d}. Unterminated list}".format(self._lineNumber))
        return None
      if token==",":
        continue
      if token=="]":
        self.PutBackToken()
        break
      self._RaiseError("Line {0:d}. Incorrect list}".format(self._lineNumber))
      return None
    token = self.GetToken()
    if token!="]":
      return None
    return managedList

  def LookUpManagedValue(self):
    if self._managedValueIndex>=len(self._managedValues):
      return None
    return self._managedValues[self._managedValueIndex]

  def GetManagedValue(self):
    if self._managedValueIndex>=len(self._managedValues):
      return None
    managedValue = self._managedValues[self._managedValueIndex]
    self._managedValueIndex += 1
    return managedValue

  def PutBackManagedValue(self):
    self._managedValueIndex -= 1

